#!/usr/bin/env python2
'''
Select values of dict of a numpy arrays, shape (1,) only based on a range
(python tuple), range is inclusive (excluding is open).

Usage:
  dictrestrict.py [options] <input> <var> <range>

Options:
  --output=OUTPUT -o OUTPUT    Output to OUTPUT instead of stdout.
  --exclude -x                 Exclude range instead of including it.    
'''
import cPickle as pickle;
import numpy as np;
from docopt import docopt;

opts = docopt(__doc__,help=True);
name = opts['<input>'];
var  = opts['<var>'];
r    = eval(opts['<range>']);
with open(name) as f:
    d=pickle.load(f);
good = (d[var] >= r[0]) & (d[var] <= r[1])
if opts['--exclude']:
    good = np.logical_not(good);
#removing
for k in d.keys():
    d[k] = d[k][good];
if opts['--output']:
    with open(opts['--output'],"wb") as f:
        pickle.dump(d,f,2);
    pass;
else:
    print(pickle.dumps(d,2));




#!/usr/bin/env python
'''
Get the first pmovie's frame's hashes  output them to a standalone file as well as the
hash specification.

Usage:
    ./firsthash.py [options] <input> <hashdout> <hashesout>

Options:
    --help -h                 Print this help.
'''
from docopt import docopt;
opts=docopt(__doc__,help=True);
import numpy as np;
from lspreader import read;
from lspreader.pmovie import firsthash;
#firsthash = firsthash_new;;
from pys import dump_pickle;
fs=read(opts['<input>'], gzip='guess');
frame1 = fs[0];
hashes, hashd  = firsthash(frame1);
hashd['gzip']='guess' #hack
hashes.sort();
np.save(opts['<hashesout>'], hashes);
dump_pickle(opts['<hashdout>'], hashd);



#!/usr/bin/env python2
'''
Turn old trajectories.npz into new trajectories.npz format.

Usage:
    ./trajnewfmt.py [options] <input> <output>

Options:
    --help -h       Print this help.
'''
from docopt import docopt;
import numpy as np;
opts=docopt(__doc__,help=True);

with np.load(opts['<input>']) as f:
    data = d['data']
    time = d['time'];
ps = data.shape[1];
out={ i:data[:,i] for i in xrange(ps) }
out['time'] = time;
np.savez(opts['<output>'],**out);



#!/usr/bin/env python2
'''
Read in pext files from a path with no restart run and output a recarray.
Requires a .lsp file to be in the path, or to be passed. If <output> is
not passed, determine the name from the .lsp name.

Usage:
   simple-pext.py [options] [<output>]

Options:
    --help -h                 This help.
    --lsp=L -L l              Read this .lsp file specifically.
    --late-time=TIME -l TIME  Cut out after this time.
    --reverse -r              Reverse Y and Z.
    --massE=ME                Rest energy of the particle. [default: 0.511e6]
    --verbose -v              Print verbose.
    --range=R -R R            Only restrict to this number range of pexts.
    --path=P -P P             Specify a path instead of the cwd. [default: .]
'''
from docopt import docopt;
from lspreader import lspreader as rd;
from lspreader.pext import add_quantities;
import numpy as np;
from pys import parse_ituple;
import re;
from lspreader.dotlsp import getdim,getpexts
import os
import numpy.lib.recfunctions as rfn;

def _vprint(s):
    print(s);
opts = docopt(__doc__,help=True);
vprint = _vprint if opts['--verbose'] else  (lambda s: None);
path = opts['--path'];

files = os.listdir(path);
pext = [f for f in files if re.search("pext[0-9]+.p4",f)];
if opts['--range']:
    a=parse_ituple(opts['--range'],length=2);
    mn,mx = min(*a),max(*a);
else:
    mn,mx = float('-inf'),float('inf');
key = [ float(re.search("pext([0-9]+).p4",f).group(1))
        for f in pext ];
pext,key = zip(*[
    (i,k) for i,k in zip(pext,key)
    if mn <= k <= mx]);
if opts['--lsp']:
    lspf=opts['--lsp'];
else:
    lspf=[f for f in files if re.search(".*\.lsp$",f)][0];
with open(lspf,"r") as f:
    lsp=f.read();
if not opts['<output>']:
    outname = re.search("(.*)\.lsp$",lspf).group(1)+"-pext";
else:
    outname = opts['<output>'];
dim=getdim(lsp);
pexts = getpexts(lsp);
latetime = float(opts['--late-time']) if opts['--late-time'] else None;
vprint('reading in files');
d = [ rd.read(name)
      for name in pext ];

d[:] = [
    rfn.rec_append_fields(
        id, 'species',
        np.ones(len(id)).astype(int)*pexts[i]['species'])
    for id,i in zip(d,key) ];
vprint('length of d={}'.format(len(d)));

d = [ i for i in d if i['t'].shape[0] > 0];
vprint('length of d={} after remove empties'.format(len(d)));
vprint('cutting out duplicate times');
if len(d) > 1:
    d = np.concatenate(d);
elif d == []:
    print("empty pext");
    quit();
else:
    d = d[0];
vprint('sorting by times')
d.sort(order='t');
if latetime:
    print('cutting out times greater than {}'.format(latetime));
    d = d[ d['t'] <= latetime ];
#calculating quantities
if opts['--reverse']:
    dim = dim[:-2] + list(reversed(dim[-2:]))
massE = float(opts['--massE']) if opts['--massE'] else None;
d = add_quantities(d, dim, massE=massE);
np.save(outname, d);



'''
For parsing .lsp files.
'''
import re;
import numpy as np;
getrx = lambda rx:float(re.search(rx,lines,flags=re.MULTILINE).group(1));

# def getextent(lsp):
#     '''
#     Obtain spatial limits of a .lsp file. Might only work with simply gridded
#     simulations.
#     '''
#     #get the grid section.
#     lines = lsp.split("\n");
#     gridstart=-1;
#     gridend=-1;
#     for i,line in enumerate(lines):
#         if gridstart >= 0:
#             if re.match("^ *\[ *Grid *\]", line):
#                 gridstart = i;
#         else:
#             if re.match("^ *\[ *[A-Z,a-z]+ *\]", line):
#                 gridend = i
#                 break;

def getdim(lsp):
    '''
    Obtain the dimensionality of a .lsp file. This should work for all well
    formatted .lsp files.

    Parameters:
    -----------
       lsp      : .lsp string

    Returns a list of dimensions.
    '''
    dims= ['x','y', 'z'];
    rxs = ['{}-cells *([0-9]+)'.format(x) for x in ['x','y','z']];
    return [
        x for x,rx in zip(dims,rxs)
        if re.search(rx,lsp) and int(re.search(rx,lsp).group(1)) > 0 ];

def getpexts(lsp):
    '''
    Get information from pext planes. This might or might not work, use with
    caution!

    Parameters:
    -----------

      lsp       : .lsp string

    Returns a list of dicts with information for all pext planes
    '''
    lines=lsp.split('\n');
    #unfortunately regex doesn't work here
    lns,planens = zip(
        *[ (i,int(re.search('^ *extract *([0-9]+)',line).group(1)))
           for i,line in enumerate(lines)
           if re.search('^ *extract *[0-9]+', line)]);
    if len(lns) == 0: return [];
    end = lns[-1];
    
    for i,line in enumerate(lines[end+1:]):
        if re.match(' *\[',line): break;
    end += i;
    lineranges = zip(lns,(lns+(end,))[1:]);
    planes=dict()
    for (i,end),plane in zip(lineranges,planens):
        d=dict();
        labels = [
            'species',
            'direction',
            'position',];
        datarx = [
            '^ *species *([0-9]+)',
            '^ *direction *([xXyYzZ])',
            '^ *at *(.*)',];
        convs  = [
            lambda s: int(s),
            lambda i: i,
            lambda s: np.array(
                map(float,s.split(' '))),
        ];
        for line in lines[i:end]:
            for label,rx,conv in zip(labels,datarx,convs):
                if re.match(rx,line):
                    d[label]=conv(re.match(rx,line).group(1));
                pass
            pass
        planes[plane] = d;
    return planes;
    
    



#!/usr/bin/env python2

'''
Interpolate something. Without an output file, interpolate in place.
There is the option of generating nearest interpolation sample files
and reusing these index files to save time on sample further files.
I intend to include this sampling in pbs scripts so that they sample
the first time step once, and then perform what I hope is a constant
time construction for later files. See the --gen-samples and --sample
options.

Usage:
  interp.py [options] <input> [<output> [<var>...]]
  interp.py [options] [--gen-samples | -G ] <input> <output>

Options:
  -h --help                   Show this help.
  -v --verbose                Turn on verbosity.
  -x --X                      Use X in interpolation.
  -y --Y                      Use Y in interpolation.
  -z --Z                      Use Z in interpolation.
  --xres=XRES                 Set the resolution along the x direction [default: 100].
  --yres=XRES                 Set the resolution along the y direction [default: 100].
  --zres=XRES                 Set the resolution along the z direction [default: 100].
  --permute -p                Swap the order of axes for 2D data.
  --gen-samples -G            Instead of interpolating, generate a sample file
                              and output to <output>.
  --sample=S -s S             Load the sample file S and use these indices to sample
                              the input file to create a interpolated file.
  --zip -z                    Compress for hdf5.
'''

from lspreader import lspreader as rd;
from lspreader.nearest import simple_nearest_indices;
import numpy as np;
import matplotlib.pyplot as plt;
from scipy.spatial import cKDTree;
from scipy.interpolate.interpnd import _ndim_coords_from_arrays,NDInterpolatorBase;
from time import time;
def handle_dims(opts):
    '''
    Script option handling.
    '''
    use,res = [],[];
    if opts['--X']:
        use.append('x');
        res.append(int(opts['--xres']));
    if opts['--Y']:
        use.append('y');
        res.append(int(opts['--yres']));
    if opts['--Z']:
        use.append('z');
        res.append(int(opts['--zres']));
    if use == []:
        use = ['x','y','z'];
        res = map(lambda k: int(opts[k]),['--xres','--yres','--zres']);
    # A couple of things to note; written in this way, whatever
    # this list (and thus, what is read) becomes, it is ordered
    # alphabetically. This is important, as this determines what
    # each resulting row and column and breadth in the output
    # array corresponds to from the actual simulation.
    #
    # It is probably worth mentioning that the xz in simulation
    # axes will be [0,1] in numpy axes, that is, it will be left-handed.
    # Using xz leads to this anyway, but it's worth reminding the reader.
    # To permute in 2D, use the --permute flag.
    return use,res;
if __name__ == "__main__":
    from docopt import docopt;
    from misc import readfile, mkvprint, dump_pickle;
    opts=docopt(__doc__,help=True);
    dims,res = handle_dims(opts);
    vprint = mkvprint;
    var = opts['<var>'];
    readvars = list(var);
    if readvars:
        readvars+=dims;
    if opts['--gen-samples']:
        xs = tuple([d[l] for l in dims]);
        i = simple_nearest_indices(xs,res);
        dump_pickle(opts["<output>"],(i,xs));
        exit(1);
    if opts['--sample']:
        i,xs = readfile(opts['--sample'], dumpfull=True);
    else:
        xs = tuple([d[l] for l in dims]);
        i = simple_nearest_indices(xs,res);
    did = {v:d[v][i] for v in var};
    #Has _D_ been _I_nterpolate_D_?  Yes it DID.
    did.update({l:x for l,x in zip(dims,xs)});
    #get it?
    #alright I'll stop
    dump_pickle(opts['<output>'], did);




#!/usr/bin/env python
'''
Read a pmovie. In the absense of an output name, output
the name based on the frame step. With an output name, and
when not outputting directly to hdf, output using dump_pickle.

Usage:
    pmov.py [options] <input> [<output>]
    pmov.py [options] [--hdf | -H ] <input> <output>

Options:
    --help -h         Output this help.
    --sort -s         Sort the pmovies by IPP.
    --hdf -H          Output to hdf5 instead of to a pickle file.
                      The group will be based on the step.
    --gzip -Z         File is gzipped.
    --zip -c          Use compression for hdf5.
    --verbose -v      Be verbose.
    --lock=L -l L     Specify a lock file for synchronized output for hdf5.
    --hash=HFILE      Hashing. It might work only for uniform
                      grids. Specify specification file used to generate hash
                      as HASHFILE.
    --firsthash=HFILE Specify first hashing, see above. Use this file as the
                      first file to generate the hash specification from and
                      output to DFILE.
    --dir=D -D D      Output to this directory if given not <output> name.
    --x -x            Use X as a spatial dimension. Similar options below are
                      for Y and Z. If none are passed, guess based on .lsp file
                      in the directory of input.
    --y -y            See above.
    --z -z            See above.
'''
from lspreader import lspreader as rd;
from lspreader.pmovie import firsthash, genhash, addhash, sortframe
from lspreader.misc import h5w;
from pys import dump_pickle, load_pickle, mkvprint;
from docopt import docopt;
import h5py as h5;
import numpy as np;
import re;

def hdfoutput(outname, frames, dozip=False):
    '''Outputs the frames to an hdf file.'''
    with h5.File(outname,'a') as f:
        for frame in frames:
            group=str(frame['step']);
            h5w(f, frame, group=group,
                compression='lzf' if dozip else None);

if __name__=='__main__':
    opts = docopt(__doc__,help=True);
    vprint = mkvprint(opts);
    #reading in using the reader.
    frames=rd.read(opts['<input>'], gzip='guess');
    if opts['--sort']:
        vprint("sorting...");
        frames[:] = [sortframe(frame) for frame in frames];
        vprint("done");
    #experimental hashing
    if opts['--firsthash'] or opts['--hash']:
        if opts['--firsthash']:
            d=firsthash(frames[0], removedupes=True);
            dump_pickle(opts['--firsthash'], d);
        else:
            d = load_pickle(opts['--hash']);
        frames[:] = [addhash(frame,d,removedupes=True) for frame in frames];
    #outputting.
    if opts['--hdf']:
        import fasteners;
        output = lambda :hdfoutput(opts['<output>'], frames, opts['--zip']);
        if opts['--lock']:
            output = fasteners.interprocess_locked(opts['--lock'])(output);
        output();
    elif not opts['<output>']:
        for frame in frames:
            outname = "{}.{}".format(opts['<input>'],frame['step']);
            if opts['--dir']:
                outname = '{}/{}'.format(opts['--dir'], outname);
            np.savez(outname, **frame);
    else:
        dump_pickle(opts['<output>'], frames);



import numpy as np;
minE=0.5;
maxX=-15e-4;
def f(frame):
    data=frame['data'];
    E   = (np.sqrt(data['ux']**2+data['uy']**2+data['uz']**2+1)-1)*0.511;
    return np.logical_and(E>minE, data['x'] < maxX);




# Copyright (c) 2015, 2016 Gregory K. Ngirmang
# All rights reserved.
#
# Copyright (C)  Pauli Virtanen, 2010.
# All rights reserved.
#
# Copyright (c) 2001, 2002 Enthought, Inc.
# All rights reserved.
#
# Copyright (c) 2003-2012 SciPy Developers.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of Enthought nor the names of the SciPy Developers
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
'''
My very own nearest interpolation for the sake of making uniform grids
'''

def nearest_indices(xs, XS):
    '''
    Returns indices that perform nearest interpolation given a
    set of points. Similar to scipy.interpolate.griddata with
    method set to "nearest". In fact, it's construction is based
    on that function and related types.

    Parameters:
    -----------

    xs -- ndarray of floats, shape (n,D)
          Data point coordinates. Can either be an array of
          shape (n,D), or a tuple of `ndim` arrays.
    XS -- ndarray of floats, shape (M, D)
          Points at which to interpolate/sample data.
    '''
    xs=_ndim_coords_from_arrays(xs);
    XS = _ndim_coords_from_arrays(XS, ndim=xs.shape[1])
    if XS.shape[-1] != xs.shape[1]:
        raise ValueError("dimensions of the points don't the sample points")
    #the magical kd-tree
    _,i = cKDTree(xs).query(XS)
    return i;

def simple_nearest_indices(xs,res):
    '''
    Simple nearest interpolator that interpolates based on
    the minima and maxima of points based on the passed
    resolution in res.

    Parameters:
    -----------

    xs  -- A collection of `ndim` arrays of points.
    res -- List of resolutions.
    '''
    maxs = [max(a) for a in xs]
    mins = [min(a) for a in xs]
    XS = [np.linspace(mn, mx, r) for mn,mx,r in zip(mins,maxs,res)];
    XS = tuple(np.meshgrid(*XS,indexing='ij'));
    if type(xs) != tuple:
        xs = tuple(xs);
    return nearest_indices(xs,XS);



'''
Useful functions for pext files.
'''
import numpy as np;
import itertools as itools;
import numpy.lib.recfunctions as rfn;

def add_quantities(d, coords=None, massE=0.511e6):
    '''
    Add physically interesting quantities to
    pext data.

    Parameters:
    -----------
      d      : pext array

    Keywords:
    ---------
      coords : sequence of positions for angle calculation. None
               or by default, calculate no angles.
               For 2D, takes the angle depending on the order passed,
               so this can be used for left-handed coordinate systems
               like LSP's xz.
      massE  : rest mass energy of particles.
    Returns a copy with the quantities appended.
    '''
    keys,items = zip(*calc_quantities(d,coords=coords,massE=massE).items());
    return rfn.rec_append_fields(
        d, keys, items);

def calc_quantities(d, coords=None, massE=0.511e6):
    '''
    Calculate physically interesting quantities from pext

    Parameters:
    -----------
      d      : pext array

    Keywords:
    ---------
      coords : sequence of positions for angle calculation. None
               or by default, calculate no angles.
               For 2D, takes the angle depending on the order passed,
               so this can be used for left-handed coordinate systems
               like LSP's xz.
      massE  : rest mass energy of particles.
    Returns a dictionary of physical quantities.
    '''
    quants = dict();
    quants['u_norm'] = np.sqrt(d['ux']**2+d['uy']**2+d['uz']**2)
    quants['KE']     =(np.sqrt(quants['u_norm']**2+1)-1)*massE;
    coords[:] = ['u'+coord for coord in coords];
    if not coords:
        pass;
    elif len(coords) == 3:
        quants['theta'] = np.arccos(d[coords[2]]/quants['u_norm']);
        quants['phi']   = np.arctan2(d[coords[1]],d[coords[0]]);
        quants['phi_n'] = np.arctan2(d[coords[2]],d[coords[0]]);
    elif len(coords) == 2:
        quants['phi']   = np.arctan2(d[coords[1]],d[coords[0]]);
    return quants;



#!/usr/bin/env python2
'''
Just get the header of a .p4 and output it.

Usage:
  p4header.py [options] <input>

Options:
  --help -h               Show this help.
  --verbose -v            Turn on verbosity.
  --size -s               Output the size of the header in bytes.
  --gzip -Z               Use gzip. Otherwise guess based on name.
'''

from lspreader import lspreader as rd;
import numpy as np;
from docopt import docopt;

opts=docopt(__doc__,help=True);
verbose = opts['--verbose'];
name = opts['<input>'];
gzip = True if opts['--gzip'] else 'guess';
h = rd.get_header(
    opts['<input>'],
    size=opts['--size'], gzip=gzip);
if opts['--size']:
    print("{}".format(h[1]));
else:
    print(h);



'''
Functions for flds and sclr files.
'''
import lspreader
from lspreader import read;
from .lspreader import get_header;
import numpy as np;
import numpy.linalg as lin;
import os;

def getvector(d,s):
    '''
    Get a vector flds data.

    Parameters:
    -----------

    d -- flds data.
    s -- key for the data.
    '''
    return np.array([d[s+"x"],d[s+"y"],d[s+"z"]]);

def vector_norm(d,k):
    '''
    Get a norm of vector flds data.

    Parameters:
    -----------

    d -- flds data.
    s -- key for the data.
    '''
    return lin.norm(getvector(d,k),axis=0)

def read_indexed(i,flds=None,sclr=None,
                 gzip='guess', dir='.', vector_norms=True,
                 keep_xs=False,gettime=False):
    '''
    A smart indexing reader that reads files by names. Looks for files
    like "<dir>/flds<i>.p4<compression>" where dir and the index are
    passed, as well as stuff from sclrs and saves them into one
    dictionary. Essentially useful for reading by timestep instead
    of by file. Assumes that both the flds and sclr are in the same
    direction.

    Parameters:
    -----------

    i -- index of file

    Required Keywords:
    ------------------

    flds  -- list of var's to load from the flds file
    sclr  -- list of var's to load from the sclr file
    
    Either one or both are required.

    Keywords:
    ---------

    gzip         -- files are gzipped. If gzip is "guess",
                    find a matching file.
    dir          -- Directory to look for files. Default is .
    vector_norms -- save the norm of the flds vectors under the
                    the name of the quantity. Default is True.
    keep_xs      -- Keep the edges. By default, False.
    gettime      -- Get the timestamp.
    '''
    fldsp4  = '{}/flds{}.p4'.format(dir,i);
    sclrp4  = '{}/sclr{}.p4'.format(dir,i);
    fldsgz  = fldsp4 + '.gz';
    sclrgz  = sclrp4 + '.gz';
    if gzip == 'guess':
        fldsname = fldsp4 if os.path.exists(fldsp4) else fldsgz
        sclrname = sclrp4 if os.path.exists(sclrp4) else sclrgz
    else:
        fldsname = fldsgz if gzip else fldsp4;
        sclrname = sclrgz if gzip else sclrp4;
    if not (flds or sclr):
        raise ValueError("Must specify flds or sclr to read.");
    elif flds is not None and sclr is not None:
        sd,srt=read(sclrname,
                    var=sclr,first_sort=True, gzip='guess',
                    keep_xs=keep_xs);
        fd=read(fldsname,
                var=flds, sort=srt, gzip='guess',
                keep_xs=keep_xs);
        ret = dict(sd=sd,fd=fd);
        ret.update({k:sd[k] for k in sd});
        ret.update({k:fd[k] for k in fd});
        if vector_norms:
            ret.update({k:vector_norm(ret,k) for k in flds})
        if gettime:
            ret['t'] = get_header(sclrname,gzip='guess')['timestamp'];
    else:
        if flds:
            var = flds;
            name= fldsname;
        else:
            var = sclr;
            name= sclrname;
        ret,_ = read(name,var=var,first_sort=True,gzip='guess');
        if flds and vector_norms:
            ret.update({k:vector_norm(ret,k) for k in flds})
        if gettime:
            ret['t'] = get_header(name,gzip='guess')['timestamp'];
    return ret;

def restrict(d,restrict):
    '''
    Restrict data by indices.

    Parameters:
    ----------

    d         -- the flds/sclr data
    restrict  -- a tuple of [xmin,xmax,...] etx
    '''
    notqs = ['t','xs','ys','zs','fd','sd']
    keys  = [k for k in d if k not in notqs];
    if len(restrict) == 2:
        for k in keys:
            d[k] = d[k][restrict[0]:restrict[1]]
    elif len(restrict) == 4:
        for k in keys:
            d[k] = d[k][
                restrict[0]:restrict[1],
                restrict[2]:restrict[3]
            ];
    elif len(restrict) == 6:
        for k in keys:
            d[k] = d[k][
                restrict[0]:restrict[1],
                restrict[2]:restrict[3],
                restrict[4]:restrict[5]
            ];
    else:
        raise ValueError("restrict of length {} is not valid".format(
            len(restrict)));





#!/usr/bin/env python2
'''
Read in pext files and output a dict. Assumes that the names
are in order of the restart files. It still isn't general enough
for any setup (which axes  you want for the angles you want
to measure phi and theta wrt). Currently, it works for angle
wrt -x for 3d and -x for 2d.

Usage:
    pext.py [options] <output> <names>...

Options:
    --X -x                    Use X.
    --Y -y                    Use Y.
    --Z -z                    Use Z.
    --late-time=TIME -l TIME  Cut out after this time.
    --reverse -r              Reverse Y and Z.
    --massE=ME                Rest energy of the particle. [default: 0.511e6]
    --verbose -v              Print verbose.
    --dict -d                 Output a dict instead of a recarray.
'''
from docopt import docopt;
from lspreader import lspreader as rd;
from lspreader.pext import add_quantities;
import numpy as np;

def _vprint(s):
    print(s);

opts = docopt(__doc__,help=True);
vprint = _vprint if opts['--verbose'] else  (lambda s: None);
outname = opts['<output>']
names = opts['<names>'];
coords = {'x':opts['--X'],'y':opts['--Y'],'z':opts['--Z']};
num_of_coords = len([i for i in coords.values() if i]);
latetime = float(opts['--late-time']) if opts['--late-time'] else None;
if num_of_coords==0:
    num_of_coords=3;
vprint('reading in files');
d = [ rd.read(name)
      for name in names ];
vprint('length of d={}'.format(len(d)));
if opts['--verbose']:
    vprint("printing d's");
    for i in d:
        vprint(i['t'].shape[0]);
d = [ i for i in d if i['t'].shape[0] > 0];
vprint('length of d={} after remove empties'.format(len(d)));
vprint('cutting out duplicate times');
#make a mask of times less than the minimum of the next pexts
#only take those in the previous run
#only assign up to the last element of d.
d[:-1] = [i[ i['t'] < j['t'].min() ] for i,j in zip(d[:-1],d[1:])]
if len(d) > 1:
    d = np.concatenate(d);
else:
    d=d[0];
if latetime:
    print('cutting out times greater than {}'.format(latetime));
    d = d[ d['t'] <= latetime ];
#calculating quantities
if num_of_coords == 2:
    #notice this only does x-y, y-z, and inverts the first.
    x = 'x' if opts['--X'] else 'y';
    y = 'y' if opts['--Y'] else 'z';
    coords = [x,y]
    if opts['--reverse']:
        coords.reverse();
elif num_of_coords ==3:
    if opts['--reverse']:
        coords = ['x','z','y'];
    else:
        coords = ['x','y','z'];

massE = float(opts['--massE']) if opts['--massE'] else None;

d = add_quantities(d, coords, massE=massE);
if opts['--dict']:
    d = {k:d[k] for k in d.dtype.names};
    from pys import dump_pickle;
    dump_pickle(outname, d);
else:
    np.save(outname, d);



'''
Reader for LSP output xdr files (.p4's)

'''
import xdrlib as xdr;
import re;
import numpy as np;
import gzip;
from pys import test;
import pys;
import sys;
if sys.version_info >= (3,0):
    strenc = lambda b: b.decode('latin1');
else:
    strenc = lambda b: b;
#get basic dtypes
def get_int(file,N=1,forcearray=False):
    ret=np.frombuffer(file.read(4*N),dtype='>i4',count=N);
    #microoptimization for GB arrays, although fromstring works easier.
    ret.flags.writeable = True
    if N==1 and not forcearray:
        return ret[0];
    return ret;

def get_float(file,N=1,forcearray=False):
    ret=np.frombuffer(file.read(4*N),dtype='>f4',count=N);
    ret.flags.writeable = True
    if N==1 and not forcearray:
        return ret[0];
    return ret;

def get_str(file):
    l1 = get_int(file);
    l2 = get_int(file);
    if l1 != l2:
        print("warning, string prefixes are not equal...");
        print("{}!={}".format(l1,l2));
    size=l1;
    if l1%4:
        size+=4-l1%4;
    #fu python3
    return strenc(xdr.Unpacker(file.read(size)).unpack_fstring(l1));

def get_list(file,fmt):
    '''makes a list out of the fmt from the LspOutput f using the format
       i for int
       f for float
       d for double
       s for string'''
    out=[]
    for i in fmt:
        if i == 'i':
            out.append(get_int(file));
        elif i == 'f' or i == 'd':
            out.append(get_float(file));
        elif i == 's':
            out.append(get_str(file));
        else:
            raise ValueError("Unexpected flag '{}'".format(i));
    return out;
def get_dict(file,fmt,keys):
    return dict(
        zip(keys, get_list(file,fmt))
    );

def get_header(file,**kw):
    '''gets the header for the .p4 file, note that this
       Advanced the file position to the end of the header.
       
       Returns the size of the header, and the size of the header,
       if the header keyword is true.
    '''
    if type(file) == str:
        #if called with a filename, recall with opened file.
        if test(kw, "gzip") and kw['gzip'] == 'guess':
            kw['gzip'] = re.search(r'\.gz$', file) is not None;
        if test(kw, "gzip"):
            with gzip.open(file,'rb') as f:
                return get_header(f,**kw);
        else:
            with open(file,'rb') as f:
                return get_header(f,**kw);
    if test(kw, "size"):
        size = file.tell();
    header = get_dict(
        file,'iiss',
        ['dump_type','dversion', 'title','revision']);
    if header['dump_type'] == 1:
        #this is a particle dump file
        d = get_dict(file,
            'fiiii',
            ['timestamp','geometry','sflagsx','sflagsy','sflagsz']);
        header.update(d);
        #reading params
        
        header['num_species'] = get_int(file);
        header['num_particles'] = get_int(file);
        nparams = get_int(file);
        units = [get_str(file) for i in range(nparams)];
        labels = ['q', 'x', 'y', 'z', 'ux', 'uy', 'uz']
        if nparams == 7:
            pass;
        elif nparams == 8 or nparams == 11:
            labels += ['H']
        elif nparams == 10 or nparams == 11:
            labels += ['xi', 'yi', 'zi'];
        else:
            raise NotImplementedError(
                'Not implemented for these number of parameters:{}.'.format(n));
        header['params'] = list(zip(labels,units));
    elif header['dump_type'] == 2 or header['dump_type'] == 3:
        #this is a fields file or a scalars file.
        d = get_dict(file,'fii',['timestamp','geometry','domains']);
        header.update(d);
        #reading quantities
        n = get_int(file);
        names=[get_str(file) for i in range(n)];
        units=[get_str(file) for i in range(n)];
        header['quantities'] = list(zip(names,units));
    elif header['dump_type'] == 6:
        #this is a particle movie file
        d = get_dict(file,
            'iiii',
            ['geometry','sflagsx','sflagsy','sflagsz']);
        header.update(d);
        #reading params
        n = get_int(file);
        flags=[bool(get_int(file)) for i in range(n)];
        units=[get_str(file) for i in range(n)];
        labels=['q','x','y','z','ux','uy','uz','E'];
        if n == 8:
            pass;
        elif n == 7:
            labels = labels[:-1];
        elif n == 11:
            labels+=['xi','yi','zi'];
        else:
            raise NotImplementedError(
                'Not implemented for these number of parameters:{}.'.format(n));
        header['params'] = [
            (label,unit) for (label,unit,flag) in zip(labels,units,flags) if flag
        ];
    elif header['dump_type'] == 10:
        #this is a particle extraction file:
        header['geometry'] = get_int(file);
        #reading quantities
        n = get_int(file);
        header['quantities'] = [get_str(file) for i in range(n)];
    else:
        raise ValueError('Unknown dump_type: {}'.format(header['dump_type']));
    if test(kw,'size'):
        return header, file.tell()-size;
    return header;


def flds_firstsort(d):
    '''
    Perform a lexsort and return the sort indices and shape as a tuple.
    '''
    shape = [ len( np.unique(d[l]) )
              for l in ['xs', 'ys', 'zs'] ];
    si = np.lexsort((d['z'],d['y'],d['x']));
    return si,shape;
def flds_sort(d,s):
    '''
    Sort based on position. Sort with s as a tuple of the sort
    indices and shape from first sort.

    Parameters:
    -----------

    d   -- the flds/sclr data
    s   -- (si, shape) sorting and shaping data from firstsort.
    '''
    labels = [ key for key in d.keys()
               if key not in ['t', 'xs', 'ys', 'zs', 'fd', 'sd'] ];
    si,shape = s;
    for l in labels:
        d[l] = d[l][si].reshape(shape);
        d[l] = np.squeeze(d[l]);
    return d;

def read_flds(file, header, var, vprint,
              vector=True,keep_edges=False,
              sort=None,first_sort=False,
              keep_xs=False,
              return_array=False):
    if vector:
        size=3;
        readin = set();
        for i in var:#we assume none of the fields end in x
            if i[-1] == 'x' or i[-1] == 'y' or i[-1] == 'z':
                readin.add(i[:-1]);
            else:
                readin.add(i);
    else:
        size=1;
        readin = set(var);
    doms = [];
    qs = [i[0] for i in header['quantities']];
    for i in range(header['domains']):
        iR, jR, kR = get_int(file, N=3);
        #getting grid parameters (real coordinates)
        nI = get_int(file); Ip = get_float(file,N=nI, forcearray=True);
        nJ = get_int(file); Jp = get_float(file,N=nJ, forcearray=True);
        nK = get_int(file); Kp = get_float(file,N=nK, forcearray=True);
        nAll = nI*nJ*nK;
        vprint('reading domain with dimensions {}x{}x{}={}.'.format(nI,nJ,nK,nAll));
        d={}
        d['xs'], d['ys'], d['zs'] = Ip, Jp, Kp;
        d['z'], d['y'], d['x'] = np.meshgrid(Kp,Jp,Ip,indexing='ij')
        d['z'], d['y'], d['x'] = d['z'].ravel(), d['y'].ravel(), d['x'].ravel();
        for quantity in qs:
            if quantity not in readin:
                vprint('skipping {}'.format(quantity));
                file.seek(nAll*4*size,1);
            else:
                vprint('reading {}'.format(quantity));
                d[quantity] = get_float(file,N=nAll*size);
                if size==3:
                    data=d[quantity].reshape(nAll,3).T;
                    d[quantity+'x'],d[quantity+'y'],d[quantity+'z']= data;
                    del data, d[quantity];
        doms.append(d);
    if not keep_edges:
        vprint("removing edges");
        dims = ['xs','ys','zs'];
        readqs = [k for k in doms[0].keys()
                  if k not in dims ] if len(doms) > 0 else None;
        mins = [ min([d[l].min() for d in doms])
                 for l in dims ];
        def cutdom(d):
            ldim = [len(d[l]) for l in dims];
            cuts = [ np.isclose(d[l][0], smin)
                     for l,smin in zip(dims,mins) ];
            cuts[:] = [None if i else 1
                       for i in cuts];
            for quantity in readqs:
                d[quantity]=d[quantity].reshape((ldim[2],ldim[1],ldim[0]));
                d[quantity]=d[quantity][cuts[2]:,cuts[1]:,cuts[0]:].ravel();
            for l,cut in zip(dims,cuts):
                d[l] = d[l][cut:];
            return d;
        doms[:] = [cutdom(d) for d in doms];
    vprint('Stringing domains together.');
    out = { k : np.concatenate([d[k] for d in doms]) for k in doms[0] };
    for k in out:
        out[k] = out[k].astype('=f4');
    if not keep_edges:
        vprint('sorting rows...');
        sort = flds_firstsort(out)
        out = flds_sort(out,sort);
    if not keep_xs:
        out.pop('xs',None);
        out.pop('ys',None);
        out.pop('zs',None);
        if return_array:
            vprint('stuffing into array'.format(k));
            keys = sorted(out.keys());
            dt = list(zip(keys,['f4']*len(out)));
            rout = np.zeros(out['x'].shape,dtype=dt);
            for k in keys:
                vprint('saving {}'.format(k));
                rout[k] = out[k];
            out=rout;
    if first_sort and not keep_edges:
        out = (out, sort);
    return out;

def iseof(file):
    c = file.tell();
    file.read(1);
    if file.tell() == c:
        return True;
    file.seek(c);
    
def read_movie(file, header):
    params,_  = zip(*header['params']);
    nparams = len(params);
    pbytes = (nparams+1)*4;
    frames=[];
    pos0 = file.tell(); 
    while not iseof(file):
        d=get_dict(file, 'fii',['t','step','pnum']);
        d['pos']=file.tell();
        file.seek(d['pnum']*pbytes,1);
        frames.append(d);
    for i,d in enumerate(frames):
        N = d['pnum'];
        dt = [('ip','>i4')]+list(zip(params,['>f4']*nparams));
        ndt= [('ip','=i4')]+list(zip(params,['=f4']*nparams));
        file.seek(d['pos']);
        arr = np.fromfile(file,dtype=dt,count=N).astype(ndt,copy=False);
        frames[i].update({'data':arr});
        del frames[i]['pos'];
    return frames;

def read_particles(file, header):
    params,_  = zip(*header['params']);
    dt = list(zip(('ip',)+params, ['>i4']+['>f4']*len(params)));
    ndt= [(i[0], re.sub(">","=",i[1])) for i in dt ];
    out = np.fromfile(file,dtype=dt,count=-1).astype(ndt,copy=False);
    return out;

def read_pext(file, header):
    nparams = len(header['quantities']);
    params = ['t','q','x','y','z','ux','uy','uz'];
    if nparams == 9:
        params+=['E'];
    elif nparams == 11:
        params+=['xi','yi','zi'];
    elif nparams == 12:
        params+=['E','xi','yi','zi'];
    #it's just floats here on out
    dt = list(zip(params, ['>f4']*len(params)));
    ndt= [ (i[0],'=f4') for  i in dt ];
    out = np.fromfile(file,dtype=dt,count=-1).astype(ndt,copy=False);
    return out;

def read(fname,**kw):
    '''
    Reads an lsp output file and returns a raw dump of data,
    sectioned into quantities either as an dictionary or a typed numpy array.

    Parameters:
    -----------

    fname -- filename of thing to read
    
    Keyword Arguments:
    ------------------

    vprint   --  Verbose printer. Used in scripts
    override --  (type, start) => A tuple of a dump type and a place to start
                 in the passed file, useful to attempting to read semicorrupted
                 files.
    gzip     --  Read as a gzip file.

    flds/sclr Specific Arguments:
    -----------------------------
    var          -- list of quantities to be read. For fields, this can consist
                    of strings that include vector components, e.g., 'Ex'. If 
                    None (default), read all quantities.
    keep_edges   -- If set to truthy, then don't remove the edges from domains before
                    concatenation and don't reshape the flds data.
    sort         -- If not None, sort using these indices, useful for avoiding
                    resorting. If True and not an ndarray, just sort.
    first_sort   -- If truthy, sort, and return the sort data for future flds
                    that should have the same shape.
    keep_xs      -- Keep the xs's, that is, the grid information. Usually redundant
                    with x,y,z returned.
    return_array -- If set to truthy, then try to return a numpy array with a dtype.
                    Requires of course that the quantities have the same shape.
    '''
    if test(kw,'gzip') and kw['gzip'] == 'guess':
        kw['gzip'] = re.search(r'\.gz$', fname) is not None;
    openf = gzip.open if test(kw, 'gzip') else open;
    with openf(fname,'rb') as file:
        if test(kw,'override'):
            dump, start = kw['override'];
            file.seek(start);
            header = {'dump_type': dump};
            if not test(kw, 'var') and 2 <= header['dump_type'] <= 3 :
                raise ValueError(
                    "If you want to force to read as a scalar, you need to supply the quantities"
                );
        else:
            header = get_header(file);
        vprint = kw['vprint'] if test(kw, 'vprint') else lambda s: None;
        if 2 <= header['dump_type'] <= 3 :
            if not test(kw, 'var'):
                var=[i[0] for i in header['quantities']];
            else:
                var=kw['var'];
            keep_edges = test(kw, 'keep_edges');
            first_sort = test(kw, 'first_sort');
            if test(kw,'sort'):
                sort = kw['sort']
            else:
                sort = None;
            keep_xs = test(kw, 'keep_xs');
            return_array = test(kw, 'return_array');
        readers = {
            1: lambda: read_particles(file, header),
            2: lambda: read_flds(
                file,header,var,vprint,
                keep_edges=keep_edges,
                first_sort=first_sort,
                sort=sort,
                keep_xs=keep_xs,
                return_array=return_array),
            3: lambda: read_flds(
                file,header,var, vprint,
                keep_edges=keep_edges,
                first_sort=first_sort,
                sort=sort,
                keep_xs=keep_xs,
                return_array=return_array,
                vector=False),
            6: lambda: read_movie(file, header),
            10:lambda: read_pext(file,header)
        };
        
        try:
            d = readers[header['dump_type']]();
        except KeyError:
            raise NotImplementedError("Other file types not implemented yet!");
    return d;

#parsing text files, taken from other things I do.
def readgridp4(gridp4,lsporder=False,str=False):
    '''
    Parse the grid.p4 text file named fname, or from a string.

    Parameters:
    -----------

    gridp4 -- filename of thing to read or in a string.
    
    Keyword Arguments:
    ------------------

    lsporder -- Return grids in lsp order as opposed to xs,ys,zs.
    str      -- Passed argument is a string, not a filename.
    
    '''
    if not str: gridp4 = pys.readtxt(gridp4);
    out = dict();
    lines = gridp4.splitlines();
    xdim = int(lines[8].strip());
    yst  = 9+xdim+1;
    ydim = int(lines[yst-1].strip());
    zst  = yst+ydim+1;
    zdim = int(lines[zst-1].strip());
    xs = np.array([float(line.strip()) for line in lines[9:9+xdim]]);
    ys = np.array([float(line.strip()) for line in lines[yst:yst+ydim]]);
    zs = np.array([float(line.strip()) for line in lines[zst:zst+zdim]]);
    if lsporder:
        return zs,ys,xs;
    else:
        return xs,ys,zs;

def readregionsp4(regionsp4,lsporder=False,str=False):
    '''
    Parse the grid.p4 text file named fname, or from a string.

    Parameters:
    -----------

    gridp4 -- filename of thing to read or in a string.
    
    Keyword Arguments:
    ------------------

    lsporder -- Return grids in lsp order as opposed to xs,ys,zs.
    str      -- Passed argument is a string, not a filename.
    
    '''
    if not str: regionsp4 = pys.readtxt(regionsp4);
    lines = regionsp4.splitlines();
    szs = lines[4::2];
    szs = [list(map(int,line.split())) for line in szs];
    lims= lines[5::2];
    lims= [list(map(float,line.split())) for line in lims];
    if lsporder:
        lims[:] = [ [r[4],r[5],r[2],r[3],r[0],r[1]] for r in lims ];
        szs[:]  = [ s[::-1] for s in szs ];
    return np.array(szs),np.array(lims);
    



#!/usr/bin/env python
'''
Extract data directly.

Usage: 
    p4read.py [options] <input> <output>

Options:
    --npy -n          Output to npy instead of pickle.
    --pickle -p       Output to pickle independent of file ending of output.
'''
from lspreader import read;
from pys import dump_pickle;
from docopt import docopt;
import numpy as np;
import re;

opts = docopt(__doc__,help=True);
if opts['--npy'] or (re.search("\.npy$",opts['<output>']) and not opts['--pickle']):
    np.save(
        opts['<output>'],
        read(opts['<input>'],return_array=True,gzip='guess'),);
else:
    dump_pickle(opts['<output>'],read(opts['<input>'],gzip='guess'));



#!/usr/bin/env python
'''
Read in pext files from directories with restart runs and output a recarray.
Requires a .lsp file to be in the path, or to be passed.

Usage:
   split-pext.py [options] <dirs>...

Options:
    --help -h                 This help.
    --lsp=L -L L              Read this .lsp file specifically.
    --late-time=TIME -l TIME  Cut out after this time.
    --reverse -r              Reverse Y and Z.
    --massE=ME                Rest energy of the particle. [default: 0.511e6]
    --verbose -v              Print verbose.
    --range=R -R R            Only restrict to this number range of pexts.
    --output=O -o O           Output to this file. Otherwise, reckon based on the
                              .lsp file name.
'''
from docopt import docopt;
from lspreader.lspreader import read_pext, get_header;
from lspreader.pext import calc_quantities;
import numpy as np;
from pys import parse_ituple, mkvprint
import re;
from lspreader.dotlsp import getdim,getpexts
import os;
import gzip;
from itertools import cycle;
import numpy.lib.recfunctions as rfn;
opts = docopt(__doc__,help=True);
vprint = mkvprint(opts);


def gzopen(*a, **kw):
    path = a[0];
    openf = gzip.open if re.search(r'\.gz$', path) else open;
    return openf(*a,**kw);

if opts['--lsp']:
    lspf=opts['--lsp'];
else:
    files = os.listdir('.');
    lspf=[f for f in files if re.search(".*\.lsp$",f)];
    if len(lspf) < 1:
        raise ValueError("Need to specify a .lsp file or see one in the same directory.");
    lspf=lspf[0];
with open(lspf,"r") as f:
    lsp=f.read();
if not opts['--output']:
    outname = re.search("(.*)\.lsp$",lspf).group(1)+"-pext";
else:
    outname = opts['--output'];
dim=getdim(lsp);
pext_info = getpexts(lsp);

if opts['--range']:
    a=parse_ituple(opts['--range'],length=2);
    mnpext,mxpext = min(*a),max(*a);
else:
    mnpext,mxpext = float('-inf'),float('inf');
outkeys = set();
allkeys = [];
def getpextfnames(path):
    files = os.listdir(path);
    pext = [f for f in files if re.search("pext[0-9]+.p4",f)];
    key = [ float(re.search("pext([0-9]+).p4",f).group(1))
            for f in pext ];
    return [ ('{}/{}'.format(path,i),k) for i,k in zip(pext,key)
             if mnpext <= k <= mxpext];
#read first directory. Must do this in order to get headers AND not skip ahead.
pextfnames = getpextfnames(opts['<dirs>'][0]);
keys = [ i[1] for i in pextfnames ];
headers = dict();
ds = dict();
for pextfname,k in pextfnames:
    with gzopen(pextfname) as f:
        header = get_header(f);
        d = read_pext(f,header);
    d = rfn.append_fields(
        d, 'species',
        np.ones(len(d)).astype(int)*pext_info[k]['species'],usemask=False);
    ds[k] = [d];
    headers[k]=header;
#obtain other directories
pextfnames = [
    (fname, headers[k], k)
    for idir in opts['<dirs>'][1:]
    for fname,k in getpextfnames(idir) ];
vprint('reading planes');
for path,header,k in pextfnames:
    with gzopen(path) as f:
        d = read_pext(f,header);
    d = rfn.append_fields(
        d, 'species',
        np.ones(len(d)).astype(int)*pext_info[k]['species'],usemask=False);
    ds[k].append(d);
vprint('stringing together');
for k in keys:
    #remove empties
    ds[k] = [ i for i in ds[k] if len(i) > 0 ];
    #make a mask of times less than the minimum of the next pexts
    #only take those in the previous run
    #only assign up to the last element of d.
    if len(ds[k]) > 1:
        ds[k][:-1] = [ i[ i['t'] < j['t'].min() ] for i,j in zip(ds[k][:-1],ds[k][1:]) ];

d = [ di for k in ds.keys() for di in ds[k] ];
vprint('read {} planes'.format(len(d)));
if len(d) == 0:
    print("no pext data");
    quit();
d = np.concatenate(d);
del ds;
latetime = float(opts['--late-time']) if opts['--late-time'] else None;
vprint('sorting by times')
d.sort(order='t');
if latetime:
    print('cutting out times greater than {}'.format(latetime));
    d = d[ d['t'] <= latetime ];
#calculating quantities
if opts['--reverse']:
    dim = dim[:-2] + list(reversed(dim[-2:]))
massE = float(opts['--massE']) if opts['--massE'] else None;
vprint('adding quantities');
qs = calc_quantities(d, coords=dim, massE=massE);
names = list(d.dtype.names) + qs.keys();
ftype = str(d.dtype[0]);
outdt = [(name,ftype) for name in names];
vprint('creating output');
out=np.empty(d.shape[0],dtype=outdt);
for k in qs:
    out[k] = qs[k];
for k in d.dtype.names:
    out[k] = d[k];
np.save(outname, out);



'''
pmovie reading functions
'''

import numpy as np;
import numpy.lib.recfunctions as rfn;
from lspreader import read;
from pys import sd, mk_getkw;
import struct;
#
# this is mainly hashing
#

def firsthash(frame, removedupes=False):
    '''
    Hashes the first time step. Only will work as long as
    the hash can fit in a uint64.

    Parameters:
    -----------
      frame : first frame.

    Keywords:
    ---------
      removedups: specify duplicates for the given frame.
    
    Returns a dictionary of everything needed
    to generate hashes from the genhash function.
    
    '''
    #hashes must have i8 available
    #overwise, we'll have overflow
    def avgdiff(d):
        d=np.sort(d);
        d = d[1:] - d[:-1]
        ret = np.average(d[np.nonzero(d)]);
        if np.isnan(ret):
            return 1.0;
        return ret;
    def hasextent(l,eps=1e-10):
        #will I one day make pic sims on the pm scale??
        dim = frame['data'][l];
        return np.abs(dim.max()-dim.min()) > eps;
    fields = list(frame['data'].dtype.names);
    dims = [ i for i in ['xi','yi','zi']
             if i in fields and hasextent(i) ];
    ip = np.array([ frame['data'][l]
                    for l in dims ]).T;
    avgdiffs = np.array([avgdiff(a) for a in ip.T]);
    mins  = ip.min(axis=0);
    ips = (((ip - mins)/avgdiffs).round().astype('uint64'))
    pws  = np.floor(np.log10(ips.max(axis=0))).astype('uint64')+1
    pws = list(pws);
    pw = [0]+[ ipw+jpw for ipw,jpw in
               zip([0]+pws[:-1],pws[:-1]) ];
    pw = 10**np.array(pw);#.astype('int64');
    #the dictionary used for hashing
    d=dict(dims=dims, mins=mins, avgdiffs=avgdiffs, pw=pw);
    hashes = genhash(frame,removedupes=False,**d);
    if removedupes:
        #consider if the negation of this is faster for genhash
        uni,counts = np.unique(hashes,return_counts=True);
        d['dupes']=uni[counts>1]
        dupei = np.in1d(hashes, d['dupes']);
        hashes[dupei] = -1;
        d['removedupes']=True;
    return hashes,d

def firsthash_new(frame,**kw):
    kw['new']=True;
    kw['dupes']=None;
    hashes = genhash(frame,**kw);
    uni,counts = np.unique(hashes,return_counts=True);
    d=sd(kw,dupes=uni[counts>1],removedupes=True);
    dupei = np.in1d(hashes, d['dupes'])
    hashes[dupei] = -1
    return hashes, retd;

genhash_defaults = dict(
    d=None,
    new=False,
    dupes=None,
    removedupes=False,
    dims=('xi','yi','zi'),
    ftype='f',
);
def genhash(frame,**kw):
    '''
    Generate the hashes for the given frame for a specification
    given in the dictionary d returned from firsthash.

    Parameters:
    -----------
      frame :  frame to hash.

    Keywords:
    ---------
      d         : hash specification generated from firsthash.
      new       : use new hashing, which isn't really hashing.
      removedups: put -1 in duplicates,
      dims      : specify dims. Supercedes the setting in `d'.
      dupes     : array of hashes known to be dupes.
      ftype     : type of floats. defaults to 'f'.

    -- old keywords from old hashing --
      mins      : minima of each axis
      avgdifs   : average differences
      pw        : powers of each axis

    Returns an array of the shape of the frames with hashes.
    '''
    getkw = mk_getkw(kw,genhash_defaults,prefer_passed=True);
    dims = getkw('dims');
    dupes= getkw('dupes');
    if not getkw('new'):
        ip = np.array([frame['data'][l] for l in dims]).T;
        scaled = ((ip - getkw('mins'))/getkw('avgdiffs')).round().astype('int64');
        hashes = (scaled*getkw('pw')).sum(axis=1).astype('int64');
    else:
        hashes = np.array([
            struct.pack('{}{}'.format(len(dims),getkw('ftype')), *[p[l] for l in dims])
            for p in frame['data']]);
    if getkw('removedupes'):
        #marking duplicated particles
        if not getkw('dupes'):
            hashes =  np.unique(hashes);
        else:
            dupei = np.in1d(hashes, getkw('dupes'));
            hashes[dupei] = -1
    return hashes;

def addhash(frame,**kw):
    '''
    helper function to add hashes to the given frame
    given in the dictionary d returned from firsthash.

    Parameters:
    -----------
      frame :  frame to hash.

    Keywords:
    ---------
      same as genhash
    
    Returns frame with added hashes, although it will be added in
    place.
    '''
    hashes = genhash(frame,**kw);
    frame['data'] = rfn.rec_append_fields(
        frame['data'],'hash',hashes);
    return frame;

def sortframe(frame):
    '''
    sorts particles for a frame
    '''
    d = frame['data'];
    sortedargs = np.lexsort([d['xi'],d['yi'],d['zi']])
    d = d[sortedargs];
    frame['data']=d;
    return frame;

def read_and_hash(fname, **kw):
    '''
    Read and and addhash each frame.
    '''
    return [addhash(frame, **kw) for frame in read(fname, **kw)];

def filter_hashes_from_file(fname, f, **kw):
    '''
    Obtain good hashes from a .p4 file with the dict hashd and a
    function that returns good hashes. Any keywords will be
    sent to read_and_hash.

    Parameters:
    -----------

    fname -- filename of file.
    f     -- function that returns a list of good hashes.
    '''
    return np.concatenate([
        frame['data']['hash'][f(frame)]
        for frame in read_and_hash(fname, **kw)
    ]);



#!/usr/bin/env python2
'''
Pointless script to concatenate numpy arrays, shape (1,) only.

Usage: arrcat.py [--help | -h] <output> <input>...
'''
import numpy as np;
from docopt import docopt;

opts = docopt(__doc__,help=True);
np.save(opts['<output>'], np.concatenate(
    tuple([np.load(name) for name in opts['<input>']])));



#!/usr/bin/env python2
'''
Make found timesteps into trajectories.

Usage:
    ./traj.py [options] <output> <input>...

Options:
    --help -h       Print this help.
    --sort -s       Sort by time.
    --new -n        New npz format, in which each particle has its own file.
'''
from docopt import docopt;
import numpy as np;
opts=docopt(__doc__,help=True);

files = opts['<input>']
def load(file):
    with np.load(file) as f:
        d,t=f['data'],f['time'][()]
    return d,t;
arrays = [load(file) for file in files];
data = np.array([arr[0] for arr in arrays]);
time = np.array([arr[1] for arr in arrays]);
if opts['--sort']:
    s=np.argsort(time);
    data=data[s];
    time=time[s];
if not opts['--new']:
    np.savez(opts['<output>'],data=data,time=time);
else:
    ps = data.shape[1];
    sz=len('{}'.format(ps));
    fmt = '{{:0{}}}'.format(sz);
    out={fmt.format(i) : data[:,i] for i in xrange(ps) };
    out['time'] = time;
    np.savez(opts['<output>'],**out);



'''
Miscellaneous shared definitions.
'''
try:
    import cPickle as pickle;
except ImportError:
    import pickle
import numpy as np;
from matplotlib import colors;
import h5py as h5;
import subprocess;

def conv(arg,default=None,func=None):
    if func:
        return func(arg) if arg else default;
    else:
        return arg if arg else default;

test = lambda d,k: k in d and d[k];

def readfile(filename, stuff='s',
             dumpfull=False,hdf=False,
             group=None):
    if hdf:
        with h5.File(filename, "r") as f:
            if group:
                g=f[group];
            else:
                g=f;
            if not stuff or dumpfull:
                d={l:np.array(g[l]) for l in g.keys()};
            elif type(stuff) == str:
                d = np.array(g[stuff]);
            else:
                d={l:np.array(g[l]) for l in stuff};
    else:
        with open(filename, "rb") as f:
            d=pickle.load(f);
        if not stuff or dumpfull:
            pass
        elif type(stuff) == str:
            d = d[stuff];
        else:
            d={l:d[l] for l in stuff};
    return d;

def dump_pickle(name, obj):
    with open(name,"wb") as f:
        pickle.dump(obj,f,2);
    pass;

def chunks(l,n):
    #http://stackoverflow.com/a/3226719
    #...not that this is hard to understand.
    return [l[x:x+n] for x in xrange(0, len(l), n)];

def check_vprint(s, vprinter):
    if vprinter is True:
        print(s);
    elif callable(vprinter):
        vprinter(s);

def mkvprint(opts):
    return lambda s: check_vprint(s,opts['--verbose']);

 
def h5w(file,d,group='/',compression=None):
    if type(file) == str:
        with h5.File(file, 'a') as f:
            h5w(f,d,group=group,
                compression=compression);
            return;
    if group not in file:
        file.create_group(group);
    group = file[group];
    for k in d:
        if k in group: del group[k];
        if type(d[k]) != np.ndarray:
            cmp = None;
        else:
            cmp = compression;
        group.create_dataset(
            k, data=d[k],  compression=cmp)
    pass;

def subcall(cmd):
    return subprocess.check_output(cmd).split('\n');

def filelines(fname,strip=False):
    with open(fname,'r') as f:
        lines = f.readlines();
    if strip:
        lines[:] = [line.strip() for line in lines]
    return lines;



#!/usr/bin/env python2
'''
Gather found particles hashes from input files.

Usage:
    ./gather.py [options] <output> <input>...

Options:
    --help -h                 Print this help.
    --intersect=I -i I        Perform intersection with this file.
'''
from docopt import docopt;
import re;
import numpy as np;


opts=docopt(__doc__,help=True);
good = set();
for fname in opts['<input>']:
    good.update(np.load(fname));
if opts['--intersect']:
    intersect = np.load(opts['--intersect']);
    good.intersection_update(good);
good = np.array(list(good));
good.sort();
np.save(opts['<output>'],good);



#!/usr/bin/env python
'''
Search a p4 for good indices. This imports the file specified by
the modulepath option, reads a function called "f" from it,
and filters the frames using it. Outputs a numpy array of
good trajectories.

Usage:
    ./search.py [options] <input> <hashd> <output>

Options:
    --help -h                    Print this help.
    --modulepath=M               Set the path to the file to read "f"
                                 from. [default: ./scanner.py]
'''
from lspreader import read;
from pys import load_pickle;
from lspreader.pmovie import filter_hashes_from_file;
import numpy as np;
import re;
import imp;
if __name__ == "__main__":
    from docopt import docopt;
    opts=docopt(__doc__,help=True);
    fname = opts['--modulepath'];
    m=re.search(r'(^.*)/(\w+)\.py$', fname);
    if not m:
        raise ValueError("module should be well named!");
    path =m.group(1);
    mname=m.group(2);
    fp, path,desc = imp.find_module(mname, [path]);
    try:
        f=imp.load_module(mname, fp, path, desc).f
    finally:
        if fp:
            fp.close();
    hashd = load_pickle(opts['<hashd>']);
    np.save(
        opts['<output>'],
        filter_hashes_from_file(opts['<input>'], f, **hashd));




from setuptools import setup

setup(name='lspreader',
      version='0.1.7r2',
      description='A set of python readers for the code LSP',
      url='http://github.com/noobermin/lspreader',
      author='noobermin',
      author_email='ngirmang.1@osu.com',
      license='MIT',
      packages=['lspreader'],
      install_requires=[
          'numpy>=1.13.1',
          'scipy>=0.19.1',
          'pys>=0.0.10',
          'docopt>=0.6.2',
      ],
      scripts=[
          'bin/p4read.py',
          'bin/split-pext.py',
          'bin/simple-pext.py',
          'bin/p4header.py'
      ],
      zip_safe=False)



#!/usr/bin/env python2
'''
Concatenate dictionaries of numpy arrays, shape (1,) only.

Usage:
  dictcat.py [options] <input>...

Options:
  -o OUTPUT --output=OUTPUT         Output to OUTPUT instead of stdout.
'''
import cPickle as pickle;
import numpy as np;
from docopt import docopt;

opts = docopt(__doc__,help=True);
names = opts['<input>'];
d=[]
for name in names:
    with open(name) as f:
        d.append(pickle.load(f));
#making into lists
out = {k:np.concatenate([i[k] for i in d]) for k in d[0].keys()};
if opts['--output']:
    with open(opts['--output'],"wb") as f:
        pickle.dump(out,f,2);
    pass;
else:
    print(pickle.dumps(out,2));




#!/usr/bin/env python2
'''
Search a p4 for the given indices.

Usage:
    ./searchp4.py [options] <input> <hashd> <indexfile>

Options:
    --help -h                    Print this help.
    --gzip -Z                    Read gzipped files.
    --dir=D -D D                 Output to this directory.
'''
from docopt import docopt;
opts=docopt(__doc__,help=True);
from pys import load_pickle;
from lspreader.pmovie import read_and_hash;
import numpy as np;

indices = np.load(opts['<indexfile>']);
hashd = load_pickle(opts['<hashd>']);
#reading in using the reader.
frames = read_and_hash(opts['<input>'], **hashd);
for frame in frames:
    data = frame['data'];
    found = np.in1d(data['hash'],indices);
    data  = data[found];
    data.sort(order='hash');
    out   = np.empty(indices.shape, dtype=data.dtype);
    out[:]      = np.nan
    out['hash'] = -1;
    out['ip']   =  0;
    outbools= np.in1d(indices, data['hash']);
    out[outbools] = data;
    outname = "_{}".format(frame['step']);
    if opts['--dir']:
        outname = '{}/{}'.format(opts['--dir'], outname);
    np.savez(outname, data=out, time=frame['t']);
